package cn.com.qmth.examcloud.core.oe.student.service.impl;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TimeZone;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.math.RandomUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.time.DateUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import com.google.common.base.Splitter;

import cn.com.qmth.examcloud.api.commons.enums.ExamSpecialSettingsType;
import cn.com.qmth.examcloud.api.commons.enums.ExamType;
import cn.com.qmth.examcloud.api.commons.security.bean.User;
import cn.com.qmth.examcloud.api.commons.security.bean.UserType;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.util.ByteUtil;
import cn.com.qmth.examcloud.commons.util.JsonUtil;
import cn.com.qmth.examcloud.commons.util.SHA256;
import cn.com.qmth.examcloud.commons.util.StringUtil;
import cn.com.qmth.examcloud.commons.util.UUID;
import cn.com.qmth.examcloud.commons.util.UrlUtil;
import cn.com.qmth.examcloud.commons.util.Util;
import cn.com.qmth.examcloud.core.oe.admin.api.ExamRecordCloudService;
import cn.com.qmth.examcloud.core.oe.admin.api.SyncExamDataCloudService;
import cn.com.qmth.examcloud.core.oe.admin.api.bean.ExamQuestionBean;
import cn.com.qmth.examcloud.core.oe.admin.api.bean.ExamRecordDataBean;
import cn.com.qmth.examcloud.core.oe.admin.api.bean.ExamRecordPaperStructBean;
import cn.com.qmth.examcloud.core.oe.admin.api.bean.ExamRecordQuestionsBean;
import cn.com.qmth.examcloud.core.oe.admin.api.request.GetPartialExamRecordDataReq;
import cn.com.qmth.examcloud.core.oe.admin.api.request.SyncExamDataReq;
import cn.com.qmth.examcloud.core.oe.admin.api.response.GetPartialExamRecordDataResp;
import cn.com.qmth.examcloud.core.oe.student.api.bean.StuExamQuestionBean;
import cn.com.qmth.examcloud.core.oe.student.api.request.CalcExamScoreReq;
import cn.com.qmth.examcloud.core.oe.student.api.request.GetExamRecordPaperStructReq;
import cn.com.qmth.examcloud.core.oe.student.api.request.GetExamRecordQuestionsReq;
import cn.com.qmth.examcloud.core.oe.student.api.response.CalcExamScoreResp;
import cn.com.qmth.examcloud.core.oe.student.api.response.GetExamRecordPaperStructResp;
import cn.com.qmth.examcloud.core.oe.student.api.response.GetExamRecordQuestionsResp;
import cn.com.qmth.examcloud.core.oe.student.base.utils.CommonUtil;
import cn.com.qmth.examcloud.core.oe.student.base.utils.QuestionTypeUtil;
import cn.com.qmth.examcloud.core.oe.student.bean.CheckExamInProgressInfo;
import cn.com.qmth.examcloud.core.oe.student.bean.CheckQrCodeInfo;
import cn.com.qmth.examcloud.core.oe.student.bean.EndExamInfo;
import cn.com.qmth.examcloud.core.oe.student.bean.GetQrCodeReq;
import cn.com.qmth.examcloud.core.oe.student.bean.GetYunSignatureReq;
import cn.com.qmth.examcloud.core.oe.student.bean.StartExamInfo;
import cn.com.qmth.examcloud.core.oe.student.bean.UploadedFileAnswerInfo;
import cn.com.qmth.examcloud.core.oe.student.bean.UpyunSignatureInfo;
import cn.com.qmth.examcloud.core.oe.student.service.ExamBossService;
import cn.com.qmth.examcloud.core.oe.student.service.ExamControlService;
import cn.com.qmth.examcloud.core.oe.student.service.ExamFaceLivenessVerifyService;
import cn.com.qmth.examcloud.core.oe.student.service.ExamRecordDataService;
import cn.com.qmth.examcloud.core.oe.student.service.ExamRecordPaperStructService;
import cn.com.qmth.examcloud.core.oe.student.service.ExamRecordQuestionsService;
import cn.com.qmth.examcloud.core.oe.student.service.ExamingSessionService;
import cn.com.qmth.examcloud.core.oe.student.service.FaceBiopsyService;
import cn.com.qmth.examcloud.core.oe.task.api.ExamCaptureCloudService;
import cn.com.qmth.examcloud.core.oe.task.api.request.SaveExamCaptureSyncCompareResultReq;
import cn.com.qmth.examcloud.core.oe.task.api.request.UpdateExamCaptureQueuePriorityReq;
import cn.com.qmth.examcloud.examwork.api.ExamCloudService;
import cn.com.qmth.examcloud.examwork.api.request.GetExamPropertyReq;
import cn.com.qmth.examcloud.examwork.api.response.GetExamPropertyResp;
import cn.com.qmth.examcloud.question.commons.core.paper.DefaultPaper;
import cn.com.qmth.examcloud.question.commons.core.paper.DefaultQuestionGroup;
import cn.com.qmth.examcloud.question.commons.core.paper.DefaultQuestionStructureWrapper;
import cn.com.qmth.examcloud.question.commons.core.paper.DefaultQuestionUnitWrapper;
import cn.com.qmth.examcloud.question.commons.core.question.QuestionType;
import cn.com.qmth.examcloud.reports.commons.bean.OnlineExamStudentReport;
import cn.com.qmth.examcloud.reports.commons.util.ReportsUtil;
import cn.com.qmth.examcloud.support.Constants;
import cn.com.qmth.examcloud.support.cache.CacheHelper;
import cn.com.qmth.examcloud.support.cache.bean.CourseCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.ExamOrgSettingsCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.ExamPropertyCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.ExamSettingsCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.ExamStudentCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.ExamStudentSettingsCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.ExtractConfigCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.ExtractConfigDetailCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.ExtractConfigPaperCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.StudentCacheBean;
import cn.com.qmth.examcloud.support.cache.bean.SysPropertyCacheBean;
import cn.com.qmth.examcloud.support.enums.ExamProperties;
import cn.com.qmth.examcloud.support.enums.ExamRecordStatus;
import cn.com.qmth.examcloud.support.enums.FaceBiopsyScheme;
import cn.com.qmth.examcloud.support.enums.HandInExamType;
import cn.com.qmth.examcloud.support.examing.ExamBoss;
import cn.com.qmth.examcloud.support.examing.ExamQuestion;
import cn.com.qmth.examcloud.support.examing.ExamRecordData;
import cn.com.qmth.examcloud.support.examing.ExamRecordPaperStruct;
import cn.com.qmth.examcloud.support.examing.ExamRecordQuestions;
import cn.com.qmth.examcloud.support.examing.ExamingActivityTime;
import cn.com.qmth.examcloud.support.examing.ExamingHeartbeat;
import cn.com.qmth.examcloud.support.examing.ExamingSession;
import cn.com.qmth.examcloud.support.examing.ExamingStatus;
import cn.com.qmth.examcloud.support.helper.ExamCacheTransferHelper;
import cn.com.qmth.examcloud.support.helper.FaceBiopsyHelper;
import cn.com.qmth.examcloud.support.redis.RedisKeyHelper;
import cn.com.qmth.examcloud.web.bootstrap.PropertyHolder;
import cn.com.qmth.examcloud.web.exception.SequenceLockException;
import cn.com.qmth.examcloud.web.helpers.SequenceLockHelper;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import cn.com.qmth.examcloud.ws.api.WsCloudService;
import cn.com.qmth.examcloud.ws.api.WsPath;
import cn.com.qmth.examcloud.ws.api.enums.WebSocketEventType;
import cn.com.qmth.examcloud.ws.api.request.SendFileAnswerMessageReq;
import cn.com.qmth.examcloud.ws.api.request.SendScanQrCodeMessageReq;
import cn.com.qmth.examcloud.ws.api.request.SendTextReq;
import main.java.com.upyun.Base64Coder;
import main.java.com.upyun.UpException;
import main.java.com.upyun.UpYunUtils;

/**
 * @author chenken
 * @date 2018年8月13日 下午2:09:08
 * @company QMTH
 * @description 在线考试控制服务实现
 */
@Service("examControlService")
public class ExamControlServiceImpl implements ExamControlService {

    private static final Logger log = LoggerFactory.getLogger(ExamControlServiceImpl.class);

    @Autowired
    private ExamRecordPaperStructService examRecordPaperStructService;

    @Autowired
    private ExamingSessionService examingSessionService;

    @Autowired
    private ExamRecordDataService examRecordDataService;

    @Autowired
    private ExamRecordQuestionsService examRecordQuestionsService;

    @Autowired
    private ExamCaptureCloudService examCaptureCloudService;

    @Autowired
    private ExamCloudService examCloudService;

    @Autowired
    private ExamBossService examBossService;

    @Autowired
    private FaceBiopsyService faceBiopsyService;

    @Autowired
    private ExamFaceLivenessVerifyService examFaceLivenessVerifyService;

    @Autowired
    private RedisClient redisClient;

    @Autowired
    private WsCloudService wsCloudService;

    @Autowired
    ExamRecordCloudService examRecordCloudService;

    @Autowired
    SyncExamDataCloudService syncExamDataCloudService;

    private static final String SEPARATOR = "/";

    private static final String UNDERLINE = "_";

    // 又拍云音频答案上传目录
    private static final String OE_ANSWER_FILE_PATH = "oe-answer-file";

    private static final String SESSION_TIMEOUT = "$core.basic.sessionTimeout";

    // 又拍云签名有效时间（秒）
    private static final Integer SIGN_TIMEOUT = 60;

    @Transactional
    @Override
    public StartExamInfo startExam(Long examStudentId, User user) {
        // 开考预处理
        prepare4Exam(examStudentId, user);

        Long studentId = user.getUserId();
        long st = System.currentTimeMillis();

        long startTime = System.currentTimeMillis();
        if (log.isDebugEnabled()) {
            log.debug("1 获取考生信息耗时:" + (System.currentTimeMillis() - startTime) + " ms");
        }
        // 检查redis session
        ExamingSession examingSession = examingSessionService.getExamingSession(studentId);
        if (examingSession == null) {
            throw new StatusException("1002", "未找到考试会话信息");
        }
        if (ExamingStatus.FORMAL.equals(examingSession.getExamingStatus())) {
            throw new StatusException("1003", "已经有考试中的科目");
        }

        // 检查并获取考试信息
        startTime = System.currentTimeMillis();
        ExamSettingsCacheBean examBean = checkExam(examingSession.getExamId(), examingSession.getStudentId());
        if (log.isDebugEnabled()) {
            log.debug("2 检查并获取考试信息耗时:" + (System.currentTimeMillis() - startTime) + " ms");
        }

        // 检查并获取课程信息
        startTime = System.currentTimeMillis();
        CourseCacheBean courseBean = checkCourse(examingSession.getCourseId());
        if (log.isDebugEnabled()) {
            log.debug("3 检查并获取课程信息耗时:" + (System.currentTimeMillis() - startTime) + " ms");
        }

        // 获取题库试卷结构（由于存在随机抽卷，所以不能缓存 ）
        startTime = System.currentTimeMillis();

        // 获取题库调卷规则
        ExtractConfigCacheBean extractConfig = CacheHelper.getExtractConfig(examingSession.getExamId(),
                courseBean.getCode());
        // 随机生成试卷
        Map<String, String> paperTypeMaps = getExamPaperByProbability(extractConfig.getDetails());
        if (paperTypeMaps.isEmpty()) {
            throw new StatusException("1006", "生成试卷失败");
        }

        String paperId = paperTypeMaps.get(examingSession.getPaperType());
        if (StringUtils.isEmpty(paperId)) {
            throw new StatusException("1007", "获取试卷失败");
        }

        // 生成试卷结构
        ExtractConfigPaperCacheBean extractConfigPaper = CacheHelper.getExtractConfigPaper(examingSession.getExamId(),
                courseBean.getCode(), examingSession.getPaperType(), paperId);
        if (log.isDebugEnabled()) {
            log.debug("5 获取题库试卷结构耗时:" + (System.currentTimeMillis() - startTime) + " ms");
        }

        // 生成考试记录
        startTime = System.currentTimeMillis();
        ExamRecordData examRecordData = examRecordDataService.createExamRecordData(examingSession, examBean, courseBean,
                paperId, extractConfigPaper.getDefaultPaper().getFullyObjective());

        // 如果开启人脸比对，将同步人脸比对结果存储到抓后结果表中
        Long rootOrgId = examRecordData.getRootOrgId();
        Long examId = examRecordData.getExamId();
        if (FaceBiopsyHelper.isFaceEnable(rootOrgId, examId, studentId)) {
            SaveExamCaptureSyncCompareResultReq req = new SaveExamCaptureSyncCompareResultReq();
            req.setExamRecordDataId(examRecordData.getId());
            req.setStudentId(user.getUserId());
            examCaptureCloudService.saveExamCaptureSyncCompareResult(req);
        }

        if (log.isDebugEnabled()) {
            log.debug("4 生成考试记录耗时:" + (System.currentTimeMillis() - startTime) + " ms");
        }


        // 小题乱序，选项乱序
        startTime = System.currentTimeMillis();
        reorderPaperStruct(extractConfig, extractConfigPaper);
        if (log.isDebugEnabled()) {
            log.debug("6 小题乱序耗时:" + (System.currentTimeMillis() - startTime) + " ms");
        }

        // 保存考试试卷结构
        startTime = System.currentTimeMillis();
        ExamRecordPaperStruct paperStruct = new ExamRecordPaperStruct();
        paperStruct.setId(UUID.randomUUID());
        paperStruct.setDefaultPaper(extractConfigPaper.getDefaultPaper());
        examRecordPaperStructService.saveExamRecordPaperStruct(examRecordData.getId(), paperStruct);
        if (log.isDebugEnabled()) {
            log.debug("7 保存考试试卷结构耗时:" + (System.currentTimeMillis() - startTime) + " ms");
        }

        // 创建考试作答记录
        startTime = System.currentTimeMillis();
        ExamRecordQuestions examRecordQuestions = examRecordQuestionsService
                .createExamRecordQuestions(examRecordData.getId(), extractConfigPaper.getDefaultPaper());
        // 记录试卷题目数量
        examRecordData.setQuestionCount(examRecordQuestions.getExamQuestions().size());
        examRecordData.setIsAllObjectivePaper(extractConfigPaper.getDefaultPaper().getFullyObjective());
        examRecordDataService.saveExamRecordDataCache(examRecordData.getId(), examRecordData);
        if (log.isDebugEnabled()) {
            log.debug("8 创建考试作答记录耗时:" + (System.currentTimeMillis() - startTime) + " ms");
        }

        // 保存考试次数控制信息
        ExamBoss eb = examBossService.getExamBoss(examingSession.getExamStudentId());
        if (eb == null) {
            throw new StatusException("008001", "ExamBoss is not created");
        }
        if (eb.getExamRecordDataIds() == null) {
            eb.setExamRecordDataIds(new ArrayList<Long>());
        }
        eb.getExamRecordDataIds().add(examRecordData.getId());
        eb.setStartCount(eb.getStartCount() + 1);
        examBossService.saveExamBoss(examStudentId, eb);

        // 初始化考试会话
        startTime = System.currentTimeMillis();
        initializeExamRecordSession(examingSession, examRecordData, examBean);
        if (log.isDebugEnabled()) {
            log.debug("9 初始化考试会话耗时:" + (System.currentTimeMillis() - startTime) + " ms");
        }

        if (log.isDebugEnabled()) {
            log.debug("10 合计 耗时:" + (System.currentTimeMillis() - st) + " ms");
        }
        // 在线考生开考打点
        ReportsUtil.report(
                new OnlineExamStudentReport(user.getRootOrgId(), user.getUserId(), examBean.getId(), examStudentId));
        StartExamInfo startExamInfo = buildStartExamInfo(examRecordData.getId(), examingSession, examBean, courseBean);
        return startExamInfo;

    }

    /**
     * 开考预处理
     *
     * @param examStudentId
     * @param user
     */
    private void prepare4Exam(Long examStudentId, User user) {

//        // 开始考试上锁,分布式锁,系统在请求结束后会，自动释放锁，无需手动解锁
//        String sequenceLockKey = Constants.EXAM_CONTROL_LOCK_PREFIX + user.getUserId();
//        SequenceLockHelper.getLock(sequenceLockKey);

        SysPropertyCacheBean stuClientLoginLimit = CacheHelper.getSysProperty("STU_CLIENT_LOGIN_LIMIT");
        Boolean stuClientLoginLimitBoolean = false;
        if (stuClientLoginLimit.getHasValue()) {
            stuClientLoginLimitBoolean = Boolean.valueOf(stuClientLoginLimit.getValue().toString());
        }
        if (stuClientLoginLimitBoolean) {
            throw new StatusException("008001", "系统维护中... ...");
        }

        ExamStudentCacheBean examStudent = CacheHelper.getExamStudent(examStudentId);

        if (null == examStudent) {
            throw new StatusException("008002", "考生不存在");
        }

        if (!examStudent.getStudentId().equals(user.getUserId().longValue())) {
            throw new StatusException("008003", "考生与当前用户不吻合");
        }

        String examingSessionKey = RedisKeyHelper.getBuilder().examingSessionKey(user.getUserId());

        ExamingSession examingSession = redisClient.get(examingSessionKey, ExamingSession.class);

        if (null != examingSession && examingSession.getExamingStatus().equals(ExamingStatus.FORMAL)) {
            throw new StatusException("008004", "已经有考试中的科目");
        }

        ExamSettingsCacheBean examSettingsCacheBean = ExamCacheTransferHelper.getCachedExam(examStudent.getExamId(),
                examStudent.getStudentId());

        StudentCacheBean studentCacheBean = CacheHelper.getStudent(examStudent.getStudentId());

        CourseCacheBean courseCacheBean = CacheHelper.getCourse(examStudent.getCourseId());
        if (!courseCacheBean.getEnable()) {
            throw new StatusException("100020", "课程被禁用");
        }

        // 如果启用了了特殊设置，并且无特殊设置时结束考试 配置 设置为true..且实际未设置特殊设置则不允许考试
        ExamPropertyCacheBean limitedIfNoSpecialSettings = ExamCacheTransferHelper.getDefaultCachedExamProperty(
                examSettingsCacheBean.getId(), ExamProperties.LIMITED_IF_NO_SPECIAL_SETTINGS.toString());

        if (examSettingsCacheBean.getSpecialSettingsEnabled() && (limitedIfNoSpecialSettings.getHasValue()
                && Boolean.valueOf(limitedIfNoSpecialSettings.getValue()))) {

            // 学生特殊设置开启未配置，不允许考试
            if (examSettingsCacheBean.getSpecialSettingsType() == ExamSpecialSettingsType.STUDENT_BASED) {
                ExamStudentSettingsCacheBean specialSettings = CacheHelper
                        .getExamStudentSettings(examSettingsCacheBean.getId(), examStudent.getStudentId());
                if (!specialSettings.getHasValue()) {
                    throw new StatusException("100014", "考试配置未完成，不允许考试");
                }
            }

            // 机构特殊设置开启未配置，不允许考试
            if (examSettingsCacheBean.getSpecialSettingsType() == ExamSpecialSettingsType.ORG_BASED) {
                // 需求调整，所有的组织机构取学生表所关联的orgId
                Long orgId = CacheHelper.getStudent(examStudent.getStudentId()).getOrgId();
                ExamOrgSettingsCacheBean specialSettings = CacheHelper.getExamOrgSettings(examSettingsCacheBean.getId(),
                        orgId);
                if (!specialSettings.getHasValue()) {
                    throw new StatusException("100015", "考试配置未完成，不允许考试");
                }
            }

        }

        if (!examSettingsCacheBean.getEnable()
                || (examSettingsCacheBean.getExamLimit() != null && examSettingsCacheBean.getExamLimit())) {
            throw new StatusException("100016", "暂无考试资格,请与学校老师联系");
        }
        if (new Date().before(examSettingsCacheBean.getBeginTime())) {
            throw new StatusException("100017", "考试未开始");
        }
        if (examSettingsCacheBean.getEndTime().before(new Date())) {
            throw new StatusException("100018", "本次考试已结束");
        }

        if ((!ExamType.ONLINE.name().equals(examSettingsCacheBean.getExamType()))
                && (!ExamType.PRACTICE.name().equals(examSettingsCacheBean.getExamType()))
                && (!ExamType.ONLINE_HOMEWORK.name().equals(examSettingsCacheBean.getExamType()))) {
            throw new StatusException("100019", "考试类型错误");
        }

        int examTimes = examSettingsCacheBean.getExamTimes().intValue();
        int usedNum = examStudent.getUsedNum();
        int extraNum = examStudent.getExtraNum();

        String examBossKey = RedisKeyHelper.getBuilder().examBossKey(examStudentId);
        ExamBoss examBoss = redisClient.get(examBossKey, ExamBoss.class);

        int asynchronousTimes = 0;
        if (null != examBoss) {
            asynchronousTimes = examBoss.getStartCount() - examBoss.getEndCount();
        } else {
            examBoss = new ExamBoss();
            examBoss.setStartCount(0);
            examBoss.setEndCount(0);
        }

        if (examTimes + extraNum <= (usedNum + asynchronousTimes)) {
            throw new StatusException("100021", "无剩余考试次数");
        }

        examingSession = new ExamingSession();
        examingSession.setCourseCode(courseCacheBean.getCode());
        examingSession.setCourseId(courseCacheBean.getId());
        examingSession.setCreationTime(new Date());
        examingSession.setExamDuration((long) examSettingsCacheBean.getDuration() * 60 * 1000);
        examingSession.setExamId(examSettingsCacheBean.getId());
        examingSession.setExamingStatus(ExamingStatus.INFORMAL);

        ExamPropertyCacheBean examReconnectTime = ExamCacheTransferHelper.getDefaultCachedExamProperty(
                examSettingsCacheBean.getId(), ExamProperties.EXAM_RECONNECT_TIME.toString());
        if (null == examReconnectTime || !examReconnectTime.getHasValue()) {
            throw new StatusException("100022", "断点续考时间未配置");
        }
        examingSession.setExamReconnectTime(StringUtil.toInteger(examReconnectTime.getValue()));

        examingSession.setExamRecordDataId(null);
        examingSession.setExamStudentId(examStudent.getExamStudentId());
        examingSession.setExamType(examSettingsCacheBean.getExamType());

        ExamPropertyCacheBean freezeTime = ExamCacheTransferHelper
                .getDefaultCachedExamProperty(examSettingsCacheBean.getId(), ExamProperties.FREEZE_TIME.toString());
        if (null == freezeTime || !freezeTime.getHasValue()) {
            throw new StatusException("100023", "交卷冻结时间未配置");
        }
        examingSession.setFreezeTime(StringUtil.toInteger(freezeTime.getValue()));
        examingSession.setOrgId(studentCacheBean.getOrgId());
        examingSession.setPaperType(examStudent.getPaperType());
        examingSession.setRootOrgId(studentCacheBean.getRootOrgId());
        examingSession.setStudentId(studentCacheBean.getId());

        redisClient.set(examingSessionKey, examingSession, 10);
        redisClient.set(examBossKey, examBoss, 60);

    }

    /**
     * 交卷
     *
     * @param examRecordDataId 考试记录id
     * @param handInExamType   交卷类型
     */
    @Transactional
    @Override
    public void handInExam(Long examRecordDataId, HandInExamType handInExamType) {
        // 此锁是为了避免自动交卷服务与断点续考交卷或活检失败交卷，同一时刻交卷争抢资源导致死锁
        String sequenceLockKey = Constants.HAND_IN_EXAM_LOCK_PREFIX + examRecordDataId;
        // 系统在请求结束后会，自动释放锁，无需手动解锁
        SequenceLockHelper.getLock(sequenceLockKey);

        ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(examRecordDataId);
        examRecordData.setHandInExamType(handInExamType);

        Long studentId = examRecordData.getStudentId();
        if (handInExamType == HandInExamType.MANUAL) {

            // 得到考试时长,校验是否达到冻结时间
            long usedExamTime = checkAndComputeExamDuration(studentId);
            examRecordData.setUsedExamTime(usedExamTime);
            examRecordData.setExamRecordStatus(ExamRecordStatus.EXAM_HAND_IN);
            examRecordData.setEndTime(new Date());

            // 手工手卷时，如果开启人脸检测，则更新抓拍队列优先级
            Long rootOrgId = examRecordData.getRootOrgId();
            Long examId = examRecordData.getExamId();

            if (FaceBiopsyHelper.isFaceEnable(rootOrgId, examId, studentId)) {
                //更新照片处理的优先级不影响交卷业务
                try {
                    // 更新照片抓拍队列优先级为高优先级
                    UpdateExamCaptureQueuePriorityReq req = new UpdateExamCaptureQueuePriorityReq();
                    req.setExamRecordDataId(examRecordDataId);
                    req.setPriority(Constants.PROCESS_CAPTURE_HIGH_PRIORITY);
                    examCaptureCloudService.updateExamCaptureQueuePriority(req);
                } catch (Exception e) {
                    log.error("[HAND_IN_EXAM-" + examRecordDataId + "]更新照片优先级时，出现异常", e);
                }
            }
        } else if (handInExamType == HandInExamType.AUTO) {
            examRecordData.setExamRecordStatus(ExamRecordStatus.EXAM_AUTO_HAND_IN);
            examRecordData.setCleanTime(new Date());
        } else {
            throw new StatusException("201101", "暂不支持的交卷类型");
        }

        //特殊处理：如果考试类型为 在线练习，则需要将部分数据提前入库，并更新相关状态
        if (ExamType.PRACTICE == examRecordData.getExamType()) {

            // 提前更新考试状态
            examRecordData.setExamRecordStatus(
                    examRecordData.getHandInExamType() == HandInExamType.MANUAL
                            ? ExamRecordStatus.EXAM_END
                            : ExamRecordStatus.EXAM_OVERDUE);

            //自动计算客观分
            CalcExamScoreReq cesReq = new CalcExamScoreReq();
            cesReq.setExamRecordDataId(examRecordDataId);
            CalcExamScoreResp calcExamScoreResp = examRecordDataService.calcExamScore(cesReq);
            examRecordData.setObjectiveScore(calcExamScoreResp.getObjectiveScore());
            examRecordData.setObjectiveAccuracy(calcExamScoreResp.getObjectiveAccuracy());
            examRecordData.setSuccPercent(calcExamScoreResp.getSuccPercent());
            examRecordData.setTotalScore(calcExamScoreResp.getTotalScore());

            //提前同步部分数据
            SyncExamDataReq syncExamDataReq = new SyncExamDataReq();
            syncExamDataReq.setExamRecordData(copyExamRecordDataFrom(examRecordData));
            syncExamDataReq.setExamRecordPaperStruct(getExamRecordPaperStruct(examRecordDataId));
            syncExamDataReq.setExamRecordQuestions(getExamRecordQuestions(examRecordDataId));
            syncExamDataCloudService.syncExamData(syncExamDataReq);
        }

        // 保存考试记录
        examRecordDataService.saveExamRecordDataCache(examRecordDataId, examRecordData);

        // 删除redis会话
        examingSessionService.deleteExamingSession(studentId);
    }

    @Override
    public EndExamInfo getEndExamInfo(Long examRecordDataId) {
        ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(examRecordDataId);
        // 如果考试没有最终结束，则返回空值
        if (null == examRecordData || (examRecordData.getExamRecordStatus() != ExamRecordStatus.EXAM_END
                && examRecordData.getExamRecordStatus() != ExamRecordStatus.EXAM_OVERDUE)) {
            return null;
        }

        //只有考试已结束 的才能返回最终结果
        if (ExamRecordStatus.EXAM_END == examRecordData.getExamRecordStatus()
                || ExamRecordStatus.EXAM_OVERDUE == examRecordData.getExamRecordStatus()) {
            EndExamInfo endExamInfo = new EndExamInfo();
            endExamInfo.setExamRecordDataId(examRecordDataId);
            endExamInfo.setIsWarn(examRecordData.getIsWarn());// 是否异常数据
            endExamInfo.setObjectiveScore(examRecordData.getObjectiveScore());// 客观题总分
            endExamInfo.setObjectiveAccuracy(examRecordData.getObjectiveAccuracy());// 客观点答对比率
            return endExamInfo;
        }

        return null;
    }

    @Override
    public CheckQrCodeInfo checkQrCode(String qrCode) {
        String str;
        str = UrlUtil.decode(qrCode);
        Map<String, String> map = Splitter.on("&").withKeyValueSeparator("=").split(str);
        String examStudentId = map.get("examStudentId");
        String examRecordDataId = map.get("examRecordDataId");
        String order = map.get("order");
        String key = map.get("key");
        String token = map.get("token");
        // 需要签名的参数
        StringBuffer sourStr = new StringBuffer();
        sourStr.append(order).append(UNDERLINE).append(examRecordDataId).append(UNDERLINE).append(key).append(UNDERLINE)
                .append(examStudentId);
        // 签名
        byte[] bytes = SHA256.encode(sourStr.toString());
        String hexAscii = ByteUtil.toHexAscii(bytes);
        if (!hexAscii.equals(token)) {
            throw new StatusException("100005", "无效的二维码");
        }

        ExamStudentCacheBean examStudent = CacheHelper.getExamStudent(Long.valueOf(examStudentId));
        if (examStudent == null) {
            throw new StatusException("100012", "考生不存在");
        }

        ExamingSession examSessionInfo = examingSessionService.getExamingSession(examStudent.getStudentId());
        String clientId;
        String userId = key.substring(key.lastIndexOf("_") + 1);

        // 未开启环境检测，才进行如下校验
        if (!isTestDev(Long.valueOf(examRecordDataId))) {
            // 非环境检测，clientId即examRecordDataId
            clientId = examRecordDataId;
            // 判断考试是否结束
            if (examSessionInfo == null) {
                throw new StatusException("100006", "考试已结束");
            }
            if (examSessionInfo.getExamRecordDataId().longValue() != Long.valueOf(examRecordDataId).longValue()
                    || examSessionInfo.getExamStudentId().longValue() != Long.valueOf(examStudentId).longValue()) {
                throw new StatusException("100008", "无效的二维码");
            }
        } else {
            // 环境检测时，clientId即用户id
            clientId = userId;
        }

        // 校验通过
        CheckQrCodeInfo res = new CheckQrCodeInfo();
        res.setExamRecordDataId(Long.valueOf(examRecordDataId));
        res.setExamStudentId(Long.valueOf(examStudentId));
        res.setKey(key);
        int sessionTimeout = PropertyHolder.getInt(SESSION_TIMEOUT, 3600);
        User user = redisClient.get(key, User.class, sessionTimeout);
        if (null == user) {
            throw new StatusException("100007", "登录信息已失效");
        }
        res.setToken(user.getToken());
        CourseCacheBean courseBean = ExamCacheTransferHelper.getCachedCourse(examStudent.getCourseId());

        res.setCourseId(courseBean.getId());
        res.setCourseName(courseBean.getName());

        ExamRecordQuestions examRecordQuestions = examRecordQuestionsService
                .getExamRecordQuestions(Long.valueOf(examRecordDataId));

        List<ExamQuestion> examQuestionList = examRecordQuestions.getExamQuestions();

        if (examRecordQuestions == null || examQuestionList == null || examQuestionList.isEmpty()) {
            throw new StatusException("100008", "无效的二维码");
        }
        List<ExamQuestion> filterList = examQuestionList.stream()
                .filter(p -> p.getOrder().equals(Integer.valueOf(order))).collect(Collectors.toList());
        if (filterList == null || filterList.isEmpty()) {
            throw new StatusException("100008", "无效的二维码");
        }
        ExamQuestion eqe = filterList.get(0);

        res.setQuestionOrder(eqe.getOrder());
        res.setQuestionMainNumber(eqe.getMainNumber());
        res.setSubNumber(getSubNumber(examRecordQuestions, Integer.valueOf(order)));
        try {
            ExamRecordData examRecordData = examRecordDataService
                    .getExamRecordDataCache(Long.valueOf(examRecordDataId));
            this.sendScanQrCodeToWebSocket(clientId, Long.valueOf(examRecordDataId), Integer.valueOf(order),
                    Long.valueOf(userId), examRecordData.getRootOrgId());
        } catch (Exception e) {
            throw new StatusException("100011", "消息通知失败", e);
        }
        return res;
    }

    /**
     * 发送消息到websocket
     *
     * @param examRecordDataId 考试记录id
     * @param order            题序号
     * @param fileUrl          文件路径
     * @param transferFileType 传输文件类型
     * @param userId
     * @throws Exception
     */
    @Override
    public void sendFileAnswerToWebSocket(Long examRecordDataId, Integer order, String fileUrl, String transferFileType,
                                          Long userId, Long rootOrgId) throws Exception {
        Map<String, Object> data = new HashMap<String, Object>();
        data.put("examRecordDataId", examRecordDataId);
        data.put("order", order);
        data.put("fileUrl", fileUrl);
        data.put("transferFileType", transferFileType);

        SendFileAnswerMessageReq sendMessageReq = new SendFileAnswerMessageReq();
        Long clientId;
        // 如果是环境检测，则使用用户id(即学生id)作为clientId
        if (isTestDev(examRecordDataId)) {
            clientId = userId;
        }
        // 不是环境检测，仍然使用考试记录id作为clientId
        else {
            clientId = examRecordDataId;
        }
        sendMessageReq.setClientId(clientId);
        sendMessageReq.setEventType(WebSocketEventType.GET_FILE_ANSWER.toString());
        sendMessageReq.setIsSuccess(true);
        sendMessageReq.setData(data);

        SendTextReq sendTextReq = new SendTextReq();
        sendTextReq.setUserType(UserType.STUDENT);
        sendTextReq.setUserId(userId);
        sendTextReq.setRootOrgId(rootOrgId);
        sendTextReq.setPath(WsPath.FILE_ANSWER);
        sendTextReq.setContent(JsonUtil.toJson(sendMessageReq));
        wsCloudService.sendText(sendTextReq);
    }

    /**
     * 通过websocket发送二维码扫描信息
     *
     * @param clientId
     * @param examRecordDataId
     * @param order
     * @throws Exception
     */
    @Override
    public void sendScanQrCodeToWebSocket(String clientId, Long examRecordDataId, Integer order, Long userId,
                                          Long rootOrgId) throws Exception {
        Map<String, Object> data = new HashMap<String, Object>();
        data.put("examRecordDataId", examRecordDataId);
        data.put("order", order);
        data.put("scanStatus", "SCANNED");
        SendScanQrCodeMessageReq sendScanQrCodeMessageReq = new SendScanQrCodeMessageReq();
        sendScanQrCodeMessageReq.setExamRecordDataId(examRecordDataId);
        sendScanQrCodeMessageReq.setEventType(WebSocketEventType.SCAN_QR_CODE.toString());
        sendScanQrCodeMessageReq.setIsSuccess(true);
        sendScanQrCodeMessageReq.setClientId(clientId);
        sendScanQrCodeMessageReq.setData(data);

        SendTextReq sendTextReq = new SendTextReq();
        sendTextReq.setUserType(UserType.STUDENT);
        sendTextReq.setUserId(userId);
        sendTextReq.setRootOrgId(rootOrgId);
        sendTextReq.setPath(WsPath.FILE_ANSWER);
        sendTextReq.setContent(JsonUtil.toJson(sendScanQrCodeMessageReq));
        wsCloudService.sendText(sendTextReq);
    }

    @Override
    public UpyunSignatureInfo getUpyunSignature(GetYunSignatureReq req) {
        UpyunSignatureInfo u = new UpyunSignatureInfo();
        try {
            ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(req.getExamRecordDataId());
            String md5 = req.getFileMd5();
            Date signDate = null;
            Date now = new Date();
            Date expirationDate = DateUtils.addSeconds(now, SIGN_TIMEOUT);

            StringBuffer filePath = new StringBuffer();

            filePath.append(SEPARATOR).append(OE_ANSWER_FILE_PATH).append(SEPARATOR)
                    .append(examRecordData.getExamStudentId()).append(SEPARATOR).append(req.getExamRecordDataId())
                    .append(SEPARATOR).append(req.getOrder()).append(SEPARATOR)
                    .append(examRecordData.getExamStudentId()).append(UNDERLINE).append(req.getExamRecordDataId())
                    .append(UNDERLINE).append(req.getOrder()).append(UNDERLINE).append(System.currentTimeMillis())
                    .append(RandomUtils.nextInt(8999) + 1000);

            if (StringUtils.isNotEmpty(req.getExt())) {
                filePath.append(UNDERLINE).append(req.getExt());
            }
            filePath.append(".").append(req.getFileSuffix());

            long expiration = expirationDate.getTime() / 1000;
            String bucketName = PropertyHolder.getString("$upyun.site.1.bucketName");
            String userName = PropertyHolder.getString("$upyun.site.1.userName");
            String password = PropertyHolder.getString("$upyun.site.1.password");

            String policy = policy(bucketName, expiration, filePath.toString(), signDate, md5);
            String sign = sign("POST", getGMTDate(signDate), bucketName, policy, userName, UpYunUtils.md5(password),
                    md5);
            u.setPolicy(policy);
            u.setSignature(sign);
            u.setFilePath(filePath.toString());

            String bucketUrl = "https://v0.api.upyun.com";
            String upyunFileUrl = PropertyHolder.getString("$upyun.site.1.domain");
            u.setUploadUrl(UrlUtil.joinUrl(bucketUrl, bucketName));
            u.setUpyunFileDomain(upyunFileUrl);
        } catch (UpException e) {
            throw new StatusException("100003", "获取又拍云签名失败");
        }
        return u;
    }

    /**
     * 获取二维码
     *
     * @param req
     * @param user
     * @return
     */
    @Override
    public String getQrCode(GetQrCodeReq req, User user) {
        if (user == null) {
            throw new StatusException("100002", "登录信息错误");
        }

        // 如果是调用环境监测的接口，不用做如下校验
        if (!req.isTestEnv()) {
            //考生id为空时，需要给考生id重新赋值
            if (req.getExamStudentId() == null) {
                ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(req.getExamRecordDataId());
                req.setExamStudentId(examRecordData.getExamStudentId());
            }

            ExamingSession examSessionInfo = examingSessionService.getExamingSession(user.getUserId());

            ExamStudentCacheBean examStudent = CacheHelper.getExamStudent(req.getExamStudentId());
            if (examStudent == null) {
                throw new StatusException("100012", "考生不存在");
            }
            if (!user.getUserId().equals(examStudent.getStudentId())) {
                throw new StatusException("100013", "无效的请求");
            }
            if (examSessionInfo == null) {
                throw new StatusException("100006", "考试已结束");
            }
            if (examSessionInfo.getExamRecordDataId().longValue() != req.getExamRecordDataId().longValue()
                    || examSessionInfo.getExamStudentId().longValue() != req.getExamStudentId().longValue()) {
                throw new StatusException("100008", "无效的请求");
            }
        }

        //环境监测时，需要特殊处理考试记录和orderId
        else {

            //考试记录id
            SysPropertyCacheBean examRecordDataIdObject = CacheHelper.getSysProperty("oe.testDev.examRecordDataId");
            //题号id
            SysPropertyCacheBean orderObject = CacheHelper.getSysProperty("oe.testDev.order");

            Long examRecordDataId;
            if (!examRecordDataIdObject.getHasValue()) {
                throw new StatusException("100014", "未找到环境监测的考试记录id");
            } else {
                examRecordDataId = Long.valueOf(examRecordDataIdObject.getValue().toString());
                req.setExamRecordDataId(examRecordDataId);
            }

            if (!orderObject.getHasValue()) {
                throw new StatusException("100015", "未找到环境监测的小题序号");
            } else {
                req.setOrder(Integer.parseInt((orderObject).getValue().toString()));
            }

            // 环境检测时，需要重新给考生id赋值
            GetPartialExamRecordDataReq gedReq = new GetPartialExamRecordDataReq();
            gedReq.setExamRecordDataId(examRecordDataId);
            GetPartialExamRecordDataResp gedResp = examRecordCloudService.getPartialExamRecordData(gedReq);

            if (null == gedResp.getExamStudentId()) {
                throw new StatusException("100016", "环境监测中配置的考试记录id不正确");
            }

            req.setExamStudentId(gedResp.getExamStudentId());
        }

        String key = user.getKey();
        StringBuffer param = new StringBuffer();
        String transferFileType = StringUtils.isBlank(req.getTransferFileType()) ? "" : req.getTransferFileType();
        param.append("examStudentId=").append(req.getExamStudentId()).append("&examRecordDataId=")
                .append(req.getExamRecordDataId()).append("&order=").append(req.getOrder()).append("&transferFileType=")
                .append(transferFileType).append("&key=").append(key);
        // 需要签名的参数
        StringBuffer sourStr = new StringBuffer();
        sourStr.append(req.getOrder()).append(UNDERLINE).append(req.getExamRecordDataId()).append(UNDERLINE).append(key)
                .append(UNDERLINE).append(req.getExamStudentId());
        // 签名
        byte[] bytes = SHA256.encode(sourStr.toString());
        String hexAscii = ByteUtil.toHexAscii(bytes);
        param.append("&token=").append(hexAscii);
        String qrStr;
        try {
            String audioAppUrl = PropertyHolder.getString("audio.app.url");
            qrStr = audioAppUrl + SEPARATOR + URLEncoder.encode(param.toString(), "UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new StatusException("100001", "参数编码异常");
        }
        return qrStr;
    }

    /**
     * @param bucketName //不能为空
     * @param expiration //不能为空
     * @param filePath   //不能为空
     * @param date       为空时，以又怕云时间和expiration比较，不为空时以此date和expiration比较
     * @param md5        //可以为空
     * @return
     */
    private String policy(String bucketName, Long expiration, String filePath, Date date, String md5) {
        Map<String, Object> paramMap = new HashMap<String, Object>();
        // 不能为空
        paramMap.put("bucket", bucketName);
        Date expirationDate = DateUtils.addSeconds(new Date(), SIGN_TIMEOUT);
        // 不能为空
        paramMap.put("expiration", expirationDate.getTime() / 1000);
        paramMap.put("save-key", filePath);
        // 为空时，以又怕云时间和expiration比较，不为空时以此date和expiration比较
        paramMap.put("date", date);
        // 可以为空
        paramMap.put("content-md5", md5);
        String policy = UpYunUtils.getPolicy(paramMap);
        return policy;
    }

    private String getGMTDate(Date d) {
        if (d == null) {
            return null;
        }
        SimpleDateFormat formater = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss 'GMT'", Locale.US);
        formater.setTimeZone(TimeZone.getTimeZone("GMT+8"));
        return formater.format(d);
    }

    /**
     * 必须和policy中date一致，可以都为空.GMTDate
     *
     * @param method
     * @param date
     * @param bucketName
     * @param policy
     * @param userName
     * @param password
     * @param md5
     * @return
     * @throws UpException
     */
    private String sign(String method, String date, String bucketName, String policy, String userName, String password,
                        String md5) throws UpException {

        StringBuilder sb = new StringBuilder();
        String sp = "&";
        sb.append(method);
        sb.append(sp);
        sb.append(SEPARATOR + bucketName);
        if (date != null) {
            sb.append(sp);
            sb.append(date);
        }
        sb.append(sp);
        sb.append(policy);
        if (md5 != null && md5.length() > 0) {
            sb.append(sp);
            sb.append(md5);
        }
        String raw = sb.toString().trim();
        byte[] hmac = null;
        try {
            hmac = UpYunUtils.calculateRFC2104HMACRaw(password, raw);
        } catch (Exception e) {
            throw new UpException("calculate SHA1 wrong.");
        }

        if (hmac != null) {
            return "UPYUN " + userName + ":" + Base64Coder.encodeLines(hmac).trim();
        }

        return null;
    }

    /**
     * 检查并返回考试 开考条件 1.enable为true 2.开始时间和结束时间判断 3.examLimit为null或false
     * 4.剩余考试次数>0
     *
     * @param examId
     * @param studentId
     * @return
     */
    private ExamSettingsCacheBean checkExam(Long examId, Long studentId) {

        // 学习中心特殊考试配置（是否禁考，开考时间可以特殊设置）
        ExamSettingsCacheBean examBean = ExamCacheTransferHelper.getCachedExam(examId, studentId);
        // 如果启用了了特殊设置，并且无特殊设置时结束考试 配置 设置为true..且实际未设置特殊设置则不允许考试
        ExamPropertyCacheBean limitedIfNoSpecialSettings = ExamCacheTransferHelper.getDefaultCachedExamProperty(examId,
                ExamProperties.LIMITED_IF_NO_SPECIAL_SETTINGS.toString());
        if (examBean.getSpecialSettingsEnabled() && (limitedIfNoSpecialSettings.getHasValue()
                && Boolean.valueOf(limitedIfNoSpecialSettings.getValue()))) {

            // 学生特殊设置开启未配置，不允许考试
            if (examBean.getSpecialSettingsType() == ExamSpecialSettingsType.STUDENT_BASED) {
                ExamStudentSettingsCacheBean specialSettings = CacheHelper.getExamStudentSettings(examId, studentId);
                if (!specialSettings.getHasValue()) {
                    throw new StatusException("2001", "考试配置未完成，不允许考试");
                }
            }

            // 机构特殊设置开启未配置，不允许考试
            if (examBean.getSpecialSettingsType() == ExamSpecialSettingsType.ORG_BASED) {
                // 需求调整，所有的组织机构取学生表所关联的orgId
                Long orgId = CacheHelper.getStudent(studentId).getOrgId();
                ExamOrgSettingsCacheBean specialSettings = CacheHelper.getExamOrgSettings(examId, orgId);
                if (!specialSettings.getHasValue()) {
                    throw new StatusException("2002", "考试配置未完成，不允许考试");
                }
            }

        }

        if (!examBean.getEnable() || (examBean.getExamLimit() != null && examBean.getExamLimit())) {
            throw new StatusException("2003", "暂无考试资格,请与学校老师联系");
        }
        if (new Date().before(examBean.getBeginTime())) {
            throw new StatusException("2004", "考试未开始");
        }
        if (examBean.getEndTime().before(new Date())) {
            throw new StatusException("2005", "本次考试已结束");
        }
        return examBean;
    }

    private CourseCacheBean checkCourse(Long courseId) {
        CourseCacheBean courseBean = ExamCacheTransferHelper.getCachedCourse(courseId);
        if (!courseBean.getEnable()) {
            throw new StatusException("3001", "该课程已被禁用");
        }
        return courseBean;
    }

    /**
     * 每个试卷类型取出一套试卷 {A:paperId,B:paperId} A是试卷类型，paperId是A类型下选定的试卷ID
     */
    private Map<String, String> getExamPaperByProbability(List<ExtractConfigDetailCacheBean> examPapers) {
        if (CollectionUtils.isEmpty(examPapers)) {
            throw new StatusException("4001", "可供抽取的试卷集合为空！");
        }

        Map<String, List<ExtractConfigDetailCacheBean>> examPaperMaps = new HashMap<>();
        for (ExtractConfigDetailCacheBean examPaper : examPapers) {
            if (examPaperMaps.containsKey(examPaper.getGroupCode())) {
                examPaperMaps.get(examPaper.getGroupCode()).add(examPaper);
            } else {
                List<ExtractConfigDetailCacheBean> list = new ArrayList<>();
                list.add(examPaper);
                examPaperMaps.put(examPaper.getGroupCode(), list);
            }
        }

        Map<String, String> paperTypeMaps = new HashMap<>();
        for (Map.Entry<String, List<ExtractConfigDetailCacheBean>> entry : examPaperMaps.entrySet()) {
            List<ExtractConfigDetailCacheBean> list = examPaperMaps.get(entry.getKey());

            String paperId = this.getPaperByProbability(list);
            if (StringUtils.isEmpty(paperId)) {
                continue;
            }

            paperTypeMaps.put(entry.getKey(), paperId);
        }

        return paperTypeMaps;
    }

    /**
     * 根据设定几率取出一套试卷
     */
    private String getPaperByProbability(List<ExtractConfigDetailCacheBean> examPapers) {
        int sum = 0;
        for (ExtractConfigDetailCacheBean examPaper : examPapers) {
            sum += examPaper.getWeight();
        }

        // 从1开始
        int r = new Random().nextInt(sum) + 1;
        for (ExtractConfigDetailCacheBean examPaper : examPapers) {
            r -= examPaper.getWeight();
            if (r <= 0) {
                return examPaper.getPaperId();// 选中
            }
        }

        return null;
    }

    /**
     * 获取试卷结构 小题乱序、选项乱序
     *
     * @param extractConfig 调卷规则对象
     * @param paperStruct   试卷结构对象
     */
    private void reorderPaperStruct(ExtractConfigCacheBean extractConfig, ExtractConfigPaperCacheBean paperStruct) {
        // 小题乱序
        if (extractConfig.getSortQuestionOrder() != null && extractConfig.getSortQuestionOrder()) {
            reorderQuestion(paperStruct.getDefaultPaper());
        }
        // 选项乱序
        if (extractConfig.getSortOptionOrder() != null && extractConfig.getSortOptionOrder()) {
            reorderOption(paperStruct.getDefaultPaper());
        }
    }

    /**
     * 小题乱序
     *
     * @param defaultPaper
     */
    private void reorderQuestion(DefaultPaper defaultPaper) {
        List<DefaultQuestionGroup> defaultQuestionGroupList = defaultPaper.getQuestionGroupList();
        for (int i = 0; i < defaultQuestionGroupList.size(); i++) {
            DefaultQuestionGroup defaultQuestionGroup = defaultQuestionGroupList.get(i);
            if (checkObjectiveQuestionByGroup(defaultQuestionGroup)) {
                List<DefaultQuestionStructureWrapper> questionStructureWrapperList = defaultQuestionGroup
                        .getQuestionWrapperList();
                Collections.shuffle(questionStructureWrapperList);
            }
        }
    }

    /**
     * 选项乱序
     *
     * @param defaultPaper
     */
    private void reorderOption(DefaultPaper defaultPaper) {
        List<DefaultQuestionGroup> defaultQuestionGroupList = defaultPaper.getQuestionGroupList();
        // 遍历大题
        for (int i = 0; i < defaultQuestionGroupList.size(); i++) {
            DefaultQuestionGroup defaultQuestionGroup = defaultQuestionGroupList.get(i);
            List<DefaultQuestionStructureWrapper> questionStructureWrapperList = defaultQuestionGroup
                    .getQuestionWrapperList();
            // 遍历小题
            for (DefaultQuestionStructureWrapper defaultQuestionStructureWrapper : questionStructureWrapperList) {
                List<DefaultQuestionUnitWrapper> questionUnitWrapperList = defaultQuestionStructureWrapper
                        .getQuestionUnitWrapperList();
                // 遍历题单元
                for (DefaultQuestionUnitWrapper defaultQuestionUnitWrapper : questionUnitWrapperList) {
                    if (QuestionTypeUtil.isChoiceQuestion(defaultQuestionUnitWrapper.getQuestionType())) {
                        Integer[] optionPermutation = defaultQuestionUnitWrapper.getOptionPermutation();
                        defaultQuestionUnitWrapper.setOptionPermutation(CommonUtil.reorderArray(optionPermutation));
                    }
                }
            }
        }
    }

    /**
     * 小题乱序条件:一个大题下全是单一的客观题，例如全是单选，全是多选，全是判断 判断一个大题下是否是单一的客观题
     *
     * @return
     */
    private boolean checkObjectiveQuestionByGroup(DefaultQuestionGroup defaultQuestionGroup) {
        Set<QuestionType> questionTypes = new HashSet<QuestionType>();
        List<DefaultQuestionStructureWrapper> questionWrapperList = defaultQuestionGroup.getQuestionWrapperList();
        for (int i = 0; i < questionWrapperList.size(); i++) {
            List<DefaultQuestionUnitWrapper> questionUnitWrapperList = questionWrapperList.get(i)
                    .getQuestionUnitWrapperList();
            if (questionUnitWrapperList.size() > 1) {
                return false;
            } else {
                questionTypes.add(questionUnitWrapperList.get(0).getQuestionType());
            }
        }
        if (questionTypes.size() > 1) {
            return false;
        }
        QuestionType questionType = questionTypes.iterator().next();
        if (QuestionTypeUtil.isObjectiveQuestion(questionType)) {
            return true;
        }
        return false;
    }

    private StartExamInfo buildStartExamInfo(Long examRecordDataId, ExamingSession examingSession,
                                             ExamSettingsCacheBean examBean, CourseCacheBean courseBean) {
        StartExamInfo startExamInfo = new StartExamInfo();
        startExamInfo.setExamRecordDataId(examRecordDataId);
        startExamInfo.setCourseName(courseBean.getName());
        startExamInfo.setDuration(examBean.getDuration());
        startExamInfo.setFaceVerifyMinute(getFaceVerifyMinute(examingSession.getRootOrgId(), examBean.getId(),
                examingSession.getOrgId(), examingSession.getStudentId()));
        return startExamInfo;
    }

    /**
     * 确定活体检测开始分钟数
     *
     * @param examId
     * @return
     */
    private Integer getFaceVerifyMinute(Long rootOrgId, Long examId, Long orgId, Long studentId) {
        // 如果开启了活体检测
        if (FaceBiopsyHelper.isFaceVerify(rootOrgId, examId, studentId)) {
            // 开始分钟数
            String startMinuteStr = ExamCacheTransferHelper
                    .getCachedExamProperty(examId, studentId, ExamProperties.FACE_VERIFY_START_MINUTE.name())
                    .getValue();
            if (CommonUtil.isBlank(startMinuteStr)) {
                throw new StatusException("5001", ExamProperties.FACE_VERIFY_START_MINUTE.getDesc() + "未设置");
            }
            Integer faceVerifyStartMinute = Integer.valueOf(startMinuteStr);

            // 结束分钟数
            String endMinuteStr = ExamCacheTransferHelper
                    .getCachedExamProperty(examId, studentId, ExamProperties.FACE_VERIFY_END_MINUTE.name()).getValue();
            if (CommonUtil.isBlank(endMinuteStr)) {
                throw new StatusException("5002", ExamProperties.FACE_VERIFY_END_MINUTE.getDesc() + "未设置");
            }
            Integer faceVerifyEndMinute = Integer.valueOf(endMinuteStr);
            return CommonUtil.calculationRandomNumber(faceVerifyStartMinute, faceVerifyEndMinute);
        }
        return null;
    }

    /**
     * 初始化考试会话
     *
     * @param examRecordData
     * @param examBean
     */
    public void initializeExamRecordSession(ExamingSession examSessionInfo, ExamRecordData examRecordData,
                                            final ExamSettingsCacheBean examBean) {
        examSessionInfo.setExamRecordDataId(examRecordData.getId());
        examSessionInfo.setStartTime(examRecordData.getStartTime().getTime());
        examSessionInfo.setExamType(examBean.getExamType());
        examSessionInfo.setExamId(examBean.getId());
        // EXAM_RECONNECT_TIME:断点续考时间
        String examReconnectTimeStr = ExamCacheTransferHelper.getCachedExamProperty(examBean.getId(),
                examSessionInfo.getStudentId(), ExamProperties.EXAM_RECONNECT_TIME.name()).getValue();
        log.debug("11.2 断点时间：" + examReconnectTimeStr);
        if (CommonUtil.isBlank(examReconnectTimeStr)) {
            throw new StatusException("6001", ExamProperties.EXAM_RECONNECT_TIME.getDesc() + "未设置");
        }
        examSessionInfo.setExamReconnectTime(Integer.valueOf(examReconnectTimeStr));
        // FREEZE_TIME:冻结时间
        String freezeTimeStr = ExamCacheTransferHelper.getCachedExamProperty(examBean.getId(),
                examSessionInfo.getStudentId(), ExamProperties.FREEZE_TIME.name()).getValue();
        log.debug("11.3 冻结时间：" + freezeTimeStr);
        if (CommonUtil.isBlank(freezeTimeStr)) {
            throw new StatusException("6002", ExamProperties.FREEZE_TIME.getDesc() + "未设置");
        }
        examSessionInfo.setFreezeTime(Integer.valueOf(freezeTimeStr));
        examSessionInfo.setExamingStatus(ExamingStatus.FORMAL);
        log.debug("11.4 开始保存考试会话...studentId=" + examSessionInfo.getStudentId());
        examingSessionService.saveExamingSession(examSessionInfo.getStudentId(), examSessionInfo);
        log.debug("11.5 保存考试会话结束 ");
    }

    @Override
    public CheckExamInProgressInfo checkExamInProgress(Long studentId) {
        ExamingSession examSessionInfo = examingSessionService.getExamingSession(studentId);
        if (examSessionInfo == null || ExamingStatus.INFORMAL.equals(examSessionInfo.getExamingStatus())) {
            return null;
        }
        // 检查考试会话是否存在，或者是否失效，如果没有失效，则返回考试中的考试记录实体，否则直接返回null
        ExamRecordData examingRecord = checkExamSession(examSessionInfo, studentId);
        if (examingRecord == null) {
            return null;
        } else {
            Integer maxInterruptNum = getMaxInterruptNum(examSessionInfo.getExamId(), examSessionInfo.getRootOrgId(),
                    examSessionInfo.getOrgId());
            CheckExamInProgressInfo checkExamInProgressInfo = new CheckExamInProgressInfo();

            if ((examingRecord.getIsExceed() == null || !examingRecord.getIsExceed())
                    && examingRecord.getContinuedCount().intValue() < maxInterruptNum.intValue()) {// 未达到最大断点次数，可继续断点一次
                // 断点续考次数自增
                int continutedCount = examingRecord.getContinuedCount() == null ? 0
                        : examingRecord.getContinuedCount().intValue();
                examingRecord.setContinuedCount(continutedCount + 1);
                examingRecord.setIsContinued(true);
                examingRecord.setIsExceed(false);
                checkExamInProgressInfo.setIsExceed(false);
            } else {
                examingRecord.setIsExceed(true);
                checkExamInProgressInfo.setIsExceed(true);
            }
            // 更新考试中的断点续考属性
            examRecordDataService.saveExamRecordDataCache(examingRecord.getId(), examingRecord);

            String examingHeartbeatKey = RedisKeyHelper.getBuilder()
                    .examingHeartbeatKey(examSessionInfo.getExamRecordDataId());
            ExamingHeartbeat examingHeartbeat = redisClient.get(examingHeartbeatKey,
                    ExamingHeartbeat.class);

            long usedTime = null == examingHeartbeat ? 0 : examingHeartbeat.getCost();

            checkExamInProgressInfo.setExamRecordDataId(examingRecord.getId());
            checkExamInProgressInfo.setExamId(examSessionInfo.getExamId());
            checkExamInProgressInfo.setUsedTime(usedTime);
            checkExamInProgressInfo.setMaxInterruptNum(maxInterruptNum);
            checkExamInProgressInfo.setInterruptNum(examingRecord.getContinuedCount());

            // 断点续考时重新计算活体检测的分钟数
            Integer faceVerifyMinute = null;
            FaceBiopsyScheme faceBiopsyScheme = FaceBiopsyHelper.getFaceBiopsyScheme(examSessionInfo.getRootOrgId());

            // 如果是新活体检测方案，则使用新的计算方案计算活检开始时间
            if (faceBiopsyScheme == FaceBiopsyScheme.NEW) {
                faceVerifyMinute = faceBiopsyService.calculateFaceBiopsyStartMinute(examingRecord.getId());
            } else {// 非新活检，默认使用旧的活检计算方式
                faceVerifyMinute = examFaceLivenessVerifyService.getFaceLivenessVerifyMinute(
                        examSessionInfo.getRootOrgId(), examSessionInfo.getOrgId(), examSessionInfo.getExamId(),
                        studentId, examSessionInfo.getExamRecordDataId(), (int) usedTime / 60);
            }

            checkExamInProgressInfo.setFaceVerifyMinute(faceVerifyMinute);
            return checkExamInProgressInfo;
        }
    }

    private ExamRecordData checkExamSession(ExamingSession examSessionInfo, Long studentId) {
        ExamRecordData examingRecord = examRecordDataService
                .getExamRecordDataCache(examSessionInfo.getExamRecordDataId());
        if (examingRecord == null) {
            return null;
        }

        String examingHeartbeatKey = RedisKeyHelper.getBuilder()
                .examingHeartbeatKey(examSessionInfo.getExamRecordDataId());
        ExamingHeartbeat examingHeartbeat = redisClient.get(examingHeartbeatKey,
                ExamingHeartbeat.class);

        long usedTime = null == examingHeartbeat ? 0 : examingHeartbeat.getCost();

        // 如果考试时间结束，自动交卷
        if (examSessionInfo.getExamDuration() <= usedTime * 1000) {
            delayHandInExamIfLocked(examingRecord.getId());
            return null;
        }

        String examingActiveTimeKey = RedisKeyHelper.getBuilder()
                .examingActiveTimeKey(examSessionInfo.getExamRecordDataId());
        ExamingActivityTime examingActiveTime = redisClient.get(examingActiveTimeKey,
                ExamingActivityTime.class);

        long activeTime = null == examingActiveTime
                ? System.currentTimeMillis()
                : examingActiveTime.getActiveTime();

        // 如果已经过了断点续考时间，自动交卷
        long now = System.currentTimeMillis();
        if (now - activeTime >= examSessionInfo.getExamReconnectTime().intValue() * 60 * 1000) {
            delayHandInExamIfLocked(examingRecord.getId());
            return null;
        }
        return examingRecord;
    }

    /**
     * 如果有序列化锁，则延迟交卷
     *
     * @param examRecordDataId 考试记录id
     * @return
     */
    private void delayHandInExamIfLocked(Long examRecordDataId) {
        try {
            handInExam(examRecordDataId, HandInExamType.AUTO);
        } catch (SequenceLockException e) {
            // 如果发现自动服务正在交卷，则重试1500毫秒获取考试记录状态，判断是否已交卷
            int loopTimes = 0;
            while (loopTimes <= 15) {
                loopTimes++;
                ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(examRecordDataId);
                if (examRecordData == null) {
                    throw new StatusException("7001", "考试信息不存在");
                }
                // 1500毫秒内如果交卷成功，则退出循环
                if (examRecordData.getExamRecordStatus() != ExamRecordStatus.EXAM_ING) {
                    return;
                }
                Util.sleep(TimeUnit.MILLISECONDS, 100);
            }
            throw e;
        }
    }

    private Integer getMaxInterruptNum(Long examId, Long rootOrgId, Long orgId) {
        GetExamPropertyReq req = new GetExamPropertyReq();
        req.setExamId(examId);
        req.setRootOrgId(rootOrgId);
        req.setOrgId(orgId);
        req.setKey(ExamProperties.MAX_INTERRUPT_NUM.name());
        GetExamPropertyResp res = examCloudService.getExamProperty(req);
        Integer ret = 100;
        if (StringUtils.isNoneBlank(res.getValue())) {
            try {
                ret = Integer.valueOf(res.getValue());
            } catch (NumberFormatException e) {
                log.error("MaxInterruptNum is not a number,return default value:100", e);
            }
        }
        return ret;
    }

    /**
     * 考试心跳每分钟调用一次
     *
     * @param user 学生
     */
    @Override
    public long examHeartbeat(User user) {
        Long studentId = user.getUserId();
        ExamingSession examSessionInfo = examingSessionService.getExamingSession(studentId);
        if (examSessionInfo == null
                || examSessionInfo.getExamingStatus().equals(ExamingStatus.INFORMAL)) {
            throw new StatusException("8001", "无效的会话,请离开考试");
        }

        String examingHeartbeatKey = RedisKeyHelper.getBuilder().examingHeartbeatKey(examSessionInfo.getExamRecordDataId());
        ExamingHeartbeat examingHeartbeat = redisClient.get(examingHeartbeatKey, ExamingHeartbeat.class);

        if (null != examingHeartbeat
                && examingHeartbeat.getCost() >= examSessionInfo.getExamDuration()) {
            throw new StatusException("8002", "考试会话已过期，请重新开考");
        }

        if (null == examingHeartbeat) {
            examingHeartbeat = new ExamingHeartbeat();
            examingHeartbeat.setCost(-60L);
            examingHeartbeat.setTimes(0L);
            examingHeartbeat.setExamRecordDataId(examSessionInfo.getExamRecordDataId());
        }

        examingHeartbeat.setTimes(examingHeartbeat.getTimes() + 1);
        examingHeartbeat.setCost(examingHeartbeat.getCost() + 60);

        redisClient.set(examingHeartbeatKey, examingHeartbeat);

        String examingActiveTimeKey = RedisKeyHelper.getBuilder()
                .examingActiveTimeKey(examSessionInfo.getExamRecordDataId());
        ExamingActivityTime examingActiveTime = redisClient.get(examingActiveTimeKey,
                ExamingActivityTime.class);
        if (null == examingActiveTime) {
            examingActiveTime = new ExamingActivityTime();
            examingActiveTime.setExamRecordDataId(examSessionInfo.getExamRecordDataId());
        }
        examingActiveTime.setActiveTime(System.currentTimeMillis());

        redisClient.set(examingActiveTimeKey, examingActiveTime);

        // 在线考生心跳打点
        ReportsUtil.report(new OnlineExamStudentReport(user.getRootOrgId(), user.getUserId(),
                examSessionInfo.getExamId(), examSessionInfo.getExamStudentId()));
        // 返回考试剩余时间
        return examSessionInfo.getExamDuration() - (examingHeartbeat.getCost() * 1000);
    }

    /**
     * 计算考试时长 校验是否达到冻结时间
     *
     * @param studentId 学生id
     * @return
     */
    private Long checkAndComputeExamDuration(Long studentId) {

        // 获取考试会话，判断考生是否已结束考试(二次校验)
        ExamingSession examingSession = examingSessionService.getExamingSession(studentId);
        if (examingSession == null) {
            throw new StatusException("oestudent-100100", "考试会话已过期");
        }

        String examingHeartbeatKey = RedisKeyHelper.getBuilder()
                .examingHeartbeatKey(examingSession.getExamRecordDataId());
        ExamingHeartbeat examingHeartbeat = redisClient.get(examingHeartbeatKey,
                ExamingHeartbeat.class);

        long cost = null == examingHeartbeat ? 0L : examingHeartbeat.getCost();

        Long examUsedMilliSeconds = cost * 1000;
        // 如果没有超过冻结时间，抛出异常
        if (ExamType.ONLINE.name().equals(examingSession.getExamType())
                || ExamType.ONLINE_HOMEWORK.name().equals(examingSession.getExamType())) {
            ExamRecordData examRecordData = examRecordDataService
                    .getExamRecordDataCache(examingSession.getExamRecordDataId());

            if (examRecordData != null && examRecordData.getIsExceed() != null && examRecordData.getIsExceed()) {// 超过断点最大次数的不校验冻结时间
                return examUsedMilliSeconds;
            }
            long freezeTime = examingSession.getFreezeTime() * 60 * 1000;
            if (examUsedMilliSeconds < freezeTime) {
                throw new StatusException("ExamControlServiceImpl-checkAndComputeExamDuration-exception",
                        "开考" + examingSession.getFreezeTime() + "分钟后才能交卷");
            }
        }

        return examUsedMilliSeconds;
    }

    private Integer getSubNumber(ExamRecordQuestions examRecordQuestions, Integer order) {
        List<UploadedFileAnswerInfo> list = getReSortedQuestionList(examRecordQuestions);
        for (UploadedFileAnswerInfo info : list) {
            if (order.intValue() == info.getOrder().intValue()) {
                return info.getSubNumber();
            }
        }
        return null;
    }

    /**
     * 根据考试记录id获取所有已经重排序号的考试列表
     *
     * @param examRecordQuestions
     * @return
     */
    private List<UploadedFileAnswerInfo> getReSortedQuestionList(ExamRecordQuestions examRecordQuestions) {
        // 所有试题集合，包括非音频题
        List<UploadedFileAnswerInfo> resultList = new ArrayList<UploadedFileAnswerInfo>();

        if (null == examRecordQuestions || null == examRecordQuestions.getExamQuestions()
                || examRecordQuestions.getExamQuestions().isEmpty()) {
            return resultList;
        }

        List<ExamQuestion> examQuestionAnswerList = examRecordQuestions.getExamQuestions();
        // 按大题号分组并排序
        Map<Integer, List<ExamQuestion>> sortedMainNumberList = examQuestionAnswerList.stream()
                .sorted(Comparator.comparing(ExamQuestion::getMainNumber))
                .collect(Collectors.groupingBy(ExamQuestion::getMainNumber, Collectors.toList()));
        for (Integer mainNum : sortedMainNumberList.keySet()) {
            // 当前大题下的小题集合
            List<ExamQuestion> subEmptyExamQuestionAnswerList = sortedMainNumberList.get(mainNum);
            // 按order对小题进行重新排序
            subEmptyExamQuestionAnswerList.sort(Comparator.comparing(ExamQuestion::getOrder));
            for (int i = 0; i < subEmptyExamQuestionAnswerList.size(); i++) {
                ExamQuestion curExamQuestion = subEmptyExamQuestionAnswerList.get(i);
                UploadedFileAnswerInfo info = new UploadedFileAnswerInfo();
                info.setExamRecordDataId(curExamQuestion.getExamRecordDataId());
                info.setMainNumber(curExamQuestion.getMainNumber());
                info.setOrder(curExamQuestion.getOrder());
                // info.setQuestionId(curExamQuestion.getQuestionId());
                info.setSubNumber(i + 1);// 每个大题下,给小题序号重新赋值
                resultList.add(info);
            }
        }
        return resultList;
    }

    /**
     * 根据考试记录id判断是否开启环境检测
     *
     * @param examRecordDataId
     * @return
     */
    private boolean isTestDev(Long examRecordDataId) {
        // 用于环境检测的考试记录id
        SysPropertyCacheBean examRecordDataIdObject = CacheHelper.getSysProperty("oe.testDev.examRecordDataId");
        // 是否开启了环境检测(请求的考试记录id等于用于环境检测的考试记录时，则认为开启了环境检测)
        return examRecordDataIdObject.getHasValue()
                && examRecordDataId.equals(Long.valueOf(examRecordDataIdObject.getValue().toString()));
    }

    private ExamRecordDataBean copyExamRecordDataFrom(ExamRecordData examRecordData) {
        ExamRecordDataBean data = new ExamRecordDataBean();
        data.setId(examRecordData.getId());
        data.setExamId(examRecordData.getExamId());
        data.setExamType(examRecordData.getExamType() == null ? null : examRecordData.getExamType().toString());
        data.setExamStudentId(examRecordData.getExamStudentId());
        data.setStudentId(examRecordData.getStudentId());
        data.setCourseId(examRecordData.getCourseId());
        data.setOrgId(examRecordData.getOrgId());
        data.setRootOrgId(examRecordData.getRootOrgId());
        data.setBasePaperId(examRecordData.getBasePaperId());
        data.setPaperType(examRecordData.getPaperType());
        data.setExamRecordStatus(examRecordData.getExamRecordStatus() == null ? null : examRecordData.getExamRecordStatus().toString());
        data.setStartTime(examRecordData.getStartTime());
        data.setEndTime(examRecordData.getEndTime());
        data.setCleanTime(examRecordData.getCleanTime());
        data.setWarn(examRecordData.getIsWarn() == null ? false : examRecordData.getIsWarn());
        data.setAudit(examRecordData.getIsAudit() == null ? false : examRecordData.getIsAudit());
        data.setIllegality(examRecordData.getIsIllegality() == null ? false : examRecordData.getIsIllegality());
        data.setUsedExamTime(examRecordData.getUsedExamTime());
        data.setContinued(examRecordData.getIsContinued() == null ? false : examRecordData.getIsContinued());
        data.setContinuedCount(examRecordData.getContinuedCount());
        data.setExceed(examRecordData.getIsExceed() == null ? false : examRecordData.getIsExceed());
        data.setFaceSuccessCount(examRecordData.getFaceSuccessCount());
        data.setFaceFailedCount(examRecordData.getFaceFailedCount());
        data.setFaceStrangerCount(examRecordData.getFaceStrangerCount());
        data.setFaceTotalCount(examRecordData.getFaceTotalCount());
        data.setFaceSuccessPercent(examRecordData.getFaceSuccessPercent());
        data.setFaceVerifyResult(examRecordData.getFaceVerifyResult() == null ? null : examRecordData.getFaceVerifyResult().toString());
        data.setBaiduFaceLivenessSuccessPercent(examRecordData.getBaiduFaceLivenessSuccessPercent());
        data.setTotalScore(examRecordData.getTotalScore());
        data.setObjectiveScore(examRecordData.getObjectiveScore());
        data.setObjectiveAccuracy(examRecordData.getObjectiveAccuracy());
        data.setSubjectiveScore(examRecordData.getSubjectiveScore());
        data.setSuccPercent(examRecordData.getSuccPercent());
        data.setAllObjectivePaper(examRecordData.getIsAllObjectivePaper());

        return data;
    }

    private ExamRecordQuestionsBean getExamRecordQuestions(Long examRecordDataId) {
        GetExamRecordQuestionsReq req = new GetExamRecordQuestionsReq();
        req.setExamRecordDataId(examRecordDataId);
        GetExamRecordQuestionsResp resp = examRecordQuestionsService.getExamRecordQuestions(req);

        ExamRecordQuestionsBean bean = new ExamRecordQuestionsBean();
        bean.setCreationTime(resp.getCreationTime());
        bean.setExamRecordDataId(resp.getExamRecordDataId());

        List<ExamQuestionBean> examQuestionBeanList = new ArrayList<>();
        for (StuExamQuestionBean stuExamQuestionBean : resp.getExamQuestions()) {
            ExamQuestionBean examQuestionBean = new ExamQuestionBean();
            examQuestionBean.setExamRecordDataId(stuExamQuestionBean.getExamRecordDataId());
            examQuestionBean.setMainNumber(stuExamQuestionBean.getMainNumber());
            examQuestionBean.setQuestionId(stuExamQuestionBean.getQuestionId());
            examQuestionBean.setOrder(stuExamQuestionBean.getOrder());
            examQuestionBean.setQuestionScore(stuExamQuestionBean.getQuestionScore());
            examQuestionBean.setQuestionType(stuExamQuestionBean.getQuestionType());
            examQuestionBean.setCorrectAnswer(stuExamQuestionBean.getCorrectAnswer());
            examQuestionBean.setStudentAnswer(stuExamQuestionBean.getStudentAnswer());
            examQuestionBean.setStudentScore(stuExamQuestionBean.getStudentScore());
            examQuestionBean.setAnswer(stuExamQuestionBean.getAnswer());
            examQuestionBean.setSign(stuExamQuestionBean.getSign());
            examQuestionBean.setOptionPermutation(stuExamQuestionBean.getOptionPermutation());
            examQuestionBean.setAudioPlayTimes(stuExamQuestionBean.getAudioPlayTimes());

            if (null != stuExamQuestionBean.getAnswerType()) {
                examQuestionBean.setAnswerType(stuExamQuestionBean.getAnswerType().name());
            }

            examQuestionBeanList.add(examQuestionBean);
        }

        bean.setExamQuestionBeans(examQuestionBeanList);

        return bean;
    }

    private ExamRecordPaperStructBean getExamRecordPaperStruct(Long examRecordDataId) {
        GetExamRecordPaperStructReq req = new GetExamRecordPaperStructReq();
        req.setExamRecordDataId(examRecordDataId);
        GetExamRecordPaperStructResp resp = examRecordPaperStructService.getExamRecordPaperStruct(req);

        ExamRecordPaperStructBean bean = new ExamRecordPaperStructBean();
        bean.setId(resp.getId());
        bean.setDefaultPaper(resp.getDefaultPaper());

        return bean;
    }
}
